#include "OptUtils.h"

#include "Timer.h"
#include "Constants.h"
#include "GzFile.h"

#include <stdio.h>
#include <stdlib.h>

#include <iostream>

/* detailed name of the trainer */
#define myname								\
  "OptUtils[" << M.name() << "]"

using namespace std;

namespace egstra {

  char* optutils::get_correct(const example* const ex) {
    const pattern* const x = ex->x();
    const int R = x->nparts();
    char* correct = (char*)calloc(R, sizeof(char));
    const label& y = ex->y();
    label::const_iterator it = y.begin();
    const label::const_iterator it_end = y.end();
    for(; it != it_end; ++it) {
      const int part = *it;
      assert(part >= 0);
      assert(part < R);
      correct[part] = 1;
    }
    return correct;
  }

  int optutils::count_matches(const example* const ex,
			      const label& yhat) {
    const int R = ex->x()->nparts();
    char* const correct = get_correct(ex);
    /* now count the number of matches */
    int match = 0;
    label::const_iterator yi = yhat.begin();
    const label::const_iterator yend = yhat.end();
    for(; yi != yend; ++yi) {
      const int part = *yi;
      assert(part >= 0);
      assert(part < R);
      match += correct[part];
    }
    free(correct);
    return match;
  }

  int optutils::hamming_loss(structured_model& M,
			     const example* const ex,
			     const label& yhat) {
    const int R = ex->x()->nparts();
    char* const lossv = M.lossvec(ex);
    /* now sum the loss indicators */
    int loss = 0;
    label::const_iterator yi = yhat.begin();
    const label::const_iterator yend = yhat.end();
    for(; yi != yend; ++yi) {
      const int part = *yi;
      assert(part >= 0);
      assert(part < R);
      loss += lossv[part];
    }
    free(lossv);
    return loss;
  }

  void optutils::classify_val_set(const parameters& V,
				  const dataset& data,
				  fgen& fg,
				  structured_model& M,
				  const string& dir,
				  const int t,
				  const bool minrisk) {
    cerr << myname << " : decoding validation set : " << flush;
    timer_start();

    /* open two output files */
    char buf[64];
    sprintf(buf, "max-labels.%03d", t);
    string maxfilename = dir + "/" + buf;
    sprintf(buf, "mbr-labels.%03d", t);
    string mbrfilename = dir + "/" + buf;
    FILE* maxout = fopen(maxfilename.c_str(), "w");
    assert(maxout != NULL);
    FILE* mbrout = NULL;
    if(minrisk) {
      mbrout = fopen(mbrfilename.c_str(), "w");
      assert(mbrout != NULL);
    }

    /* use these to compute the part-wise precision and recall */
    int correct_parts = 0;
    int max_matching_parts = 0;
    int max_predicted_parts = 0;
    int mbr_matching_parts = 0;
    int mbr_predicted_parts = 0;
    int hamming_tot = 0;
    int max_hamming_sum = 0;
    int mbr_hamming_sum = 0;

    const int N = data.size();
    int cnt = 0;
    dataset::const_iterator it = data.begin();
    const dataset::const_iterator it_end = data.end();
    for(; it != it_end; ++it) {
      const example* ex = *it;
      const pattern* const x = ex->x();

      /* compute inner products */
      const int R = x->nparts();
      double* const S = (double*)malloc(R*sizeof(double));
      const struct fvec* F = fg.phi(x);
      V.dot(S, F, R);
      fg.discard(F, x);

      const label& y = ex->y();
      correct_parts += y.size();
      hamming_tot += ex->x()->size();

      /* find the argmax (viterbi) label */
      label ymax;
      M.argmax(x, S, ymax);
      M.fwrite(maxout, x, ymax);
      max_predicted_parts += ymax.size();
      max_matching_parts += count_matches(ex, ymax);
      max_hamming_sum += hamming_loss(M, ex, ymax);

      /* find the min-risk label */
      if(minrisk) {
	double* const P = (double*)malloc(R*sizeof(double));
	M.marginals(x, S, P);
	label ymbr;
	M.argmax(x, P, ymbr);
	free(P);
	M.fwrite(mbrout, x, ymbr);
	mbr_predicted_parts += ymbr.size();
	mbr_matching_parts += count_matches(ex, ymbr);
	mbr_hamming_sum += hamming_loss(M, ex, ymbr);
      }

      free(S);

      ++cnt;
      ticker(cnt, N);
    }
    ticker(N, N);
    fclose(maxout);
    if(minrisk) { fclose(mbrout); }
    timer_stop();
    cerr << " done [" << timestr(timediff) << "]" << endl;
    assert(hamming_tot > 0);
    const double maxH = 100.0 - (100.0*(double)max_hamming_sum/hamming_tot);
    cerr << myname << " : argmax hamming score = "
	 << fixed << setprecision(2) << maxH << endl;
    assert(correct_parts > 0);
    assert(max_predicted_parts > 0);
    const double maxP =
      100.0*(double)max_matching_parts/(double)max_predicted_parts;
    const double maxR =
      100.0*(double)max_matching_parts/(double)correct_parts;
    cerr << myname << " : argmax prec/rec = "
	 << fixed << setprecision(2)
	 << maxP << " " << maxR << endl;
    if(minrisk) {
      const double mbrH = 100.0 - (100.0*(double)mbr_hamming_sum/hamming_tot);
      cerr << myname << " : min-risk hamming score = "
	   << fixed << setprecision(2) << mbrH << endl;
      assert(mbr_predicted_parts > 0);
      const double mbrP =
	100.0*(double)mbr_matching_parts/(double)mbr_predicted_parts;
      const double mbrR =
	100.0*(double)mbr_matching_parts/(double)correct_parts;
      cerr << myname << " : min-risk prec/rec = "
	   << fixed << setprecision(2)
	   << mbrP << " " << mbrR << endl;
    }
  }


  void optutils::kclassify_val_set(const parameters& V,
				   const dataset& data,
				   fgen& fg,
				   structured_model& M,
				   const string& dir,
				   const int t,
				   const int K,
				   const bool minrisk) {
    cerr << myname << " : k-best decoding validation set : " << flush;
    timer_start();

    /* open two output files */
    char buf[64];
    sprintf(buf, "kmax-labels.%03d.gz", t);
    string maxfilename = dir + "/" + buf;
    FILE* maxout = gzfile::gzopen(maxfilename.c_str(), "w");
    sprintf(buf, "kmax-labels.%03d.scores.gz", t);
    string maxscfilename = dir + "/" + buf;
    FILE* maxscout = gzfile::gzopen(maxscfilename.c_str(), "w");
    string mbrfilename;
    string mbrscfilename;
    FILE* mbrout = NULL;
    FILE* mbrscout = NULL;
    if(minrisk) {
      sprintf(buf, "kmbr-labels.%03d.gz", t);
      mbrfilename = dir + "/" + buf;
      mbrout = gzfile::gzopen(mbrfilename.c_str(), "w");
      sprintf(buf, "kmbr-labels.%03d.scores.gz", t);
      mbrscfilename = dir + "/" + buf;
      mbrscout = gzfile::gzopen(mbrscfilename.c_str(), "w");
    }

    /* use these to compute the part-wise precision and recall */
    int correct_parts = 0;
    int max_matching_parts = 0;
    int max_predicted_parts = 0;
    int max_union_recall = 0;
    int mbr_matching_parts = 0;
    int mbr_predicted_parts = 0;
    int mbr_union_recall = 0;
    int hamming_tot = 0;
    int max_hamming_sum = 0;
    int mbr_hamming_sum = 0;

    const int N = data.size();
    int cnt = 0;
    dataset::const_iterator it = data.begin();
    const dataset::const_iterator it_end = data.end();
    for(; it != it_end; ++it) {
      const example* ex = *it;
      const pattern* const x = ex->x();

      /* compute inner products */
      const int R = x->nparts();
      double* const S = (double*)malloc(R*sizeof(double));
      const struct fvec* F = fg.phi(x);
      V.dot(S, F, R);
      fg.discard(F, x);

      const label& y = ex->y();
      correct_parts += y.size();
      hamming_tot += ex->x()->size();
      double score;

      /* find the k-best labels */
      int best_hamming = 0;
      int best_match = 0;
      int best_parts = 0;
      char* flags = get_correct(ex);
      M.kbest_init(x, S);
      for(int k = 1; k <= K; ++k) {
	label ymax;
	if(!M.kbest(k, x, S, ymax, score)) { break; }
	M.fwrite(maxout, x, ymax);
	fprintf(maxscout, "%d %d %.16g\n", x->id(), k - 1, score);
	/* best #matching parts (for precision/recall) */
	const int match = count_matches(ex, ymax);
	if(k == 1 || best_match < match) {
	  best_match = match;
	  best_parts = ymax.size();
	}
	/* lowest hamming loss */
	const int hamming = hamming_loss(M, ex, ymax);
	if(k == 1 || best_hamming > hamming) {
	  best_hamming = hamming;
	}
	/* compute the recall of the union of all k-best parts */
	label::const_iterator it = ymax.begin();
	const label::const_iterator it_end = ymax.end();
	for(; it != it_end; ++it) {
	  const int r = *it;
	  assert(r >= 0);
	  assert(r < R);
	  if(flags[r]) {
	    flags[r] = 0;
	    ++max_union_recall;
	  }
	}
      }
      M.kbest_clear();
      max_matching_parts  += best_match;
      max_predicted_parts += best_parts;
      max_hamming_sum     += best_hamming;
      free(flags);

      /* find the min-risk label */
      if(minrisk) {
	double* const P = (double*)malloc(R*sizeof(double));
	M.marginals(x, S, P);
	best_match = 0;
	best_parts = 0;
	flags = get_correct(ex);
	M.kbest_init(x, P);
	for(int k = 1; k <= K; ++k) {
	  label ymbr;
	  if(!M.kbest(k, x, P, ymbr, score)) { break; }
	  M.fwrite(mbrout, x, ymbr);
	  fprintf(mbrscout, "%d %d %.16g\n", x->id(), k - 1, score);
	  /* best #matching parts (for precision/recall) */
	  const int match = count_matches(ex, ymbr);
	  if(k == 1 || best_match < match) {
	    best_match = match;
	    best_parts = ymbr.size();
	  }
	  /* lowest hamming loss */
	  const int hamming = hamming_loss(M, ex, ymbr);
	  if(k == 1 || best_hamming > hamming) {
	    best_hamming = hamming;
	  }
	  /* compute the recall of the union of all k-best parts */
	  label::const_iterator it = ymbr.begin();
	  const label::const_iterator it_end = ymbr.end();
	  for(; it != it_end; ++it) {
	    const int r = *it;
	    assert(r >= 0);
	    assert(r < R);
	    if(flags[r]) {
	      flags[r] = 0;
	      ++mbr_union_recall;
	    }
	  }
	}
	M.kbest_clear();
	mbr_matching_parts  += best_match;
	mbr_predicted_parts += best_parts;
	mbr_hamming_sum     += best_hamming;
	free(P);
	free(flags);
      }

      free(S);

      ++cnt;
      ticker(cnt, N);
    }
    ticker(N, N);
    /* close the output files */
    gzfile::gzclose(maxfilename.c_str(),   maxout);
    gzfile::gzclose(maxscfilename.c_str(), maxscout);
    if(minrisk) {
      gzfile::gzclose(mbrfilename.c_str(),   mbrout);
      gzfile::gzclose(mbrscfilename.c_str(), mbrscout);
    }
    timer_stop();
    cerr << " done [" << timestr(timediff) << "]" << endl;
    assert(hamming_tot > 0);
    const double maxH = 100.0 - (100.0*(double)max_hamming_sum/hamming_tot);
    cerr << myname << " : argmax oracle hamming score = "
	 << fixed << setprecision(2) << maxH << endl;
    assert(correct_parts > 0);
    assert(max_predicted_parts > 0);
    const double maxP =
      100.0*(double)max_matching_parts/(double)max_predicted_parts;
    const double maxR =
      100.0*(double)max_matching_parts/(double)correct_parts;
    const double maxU =
      100.0*(double)max_union_recall/(double)correct_parts;
    cerr << myname << " : argmax oracle prec/rec = "
	 << fixed << setprecision(2)
	 << maxP << " " << maxR << endl;
    cerr << myname << " : argmax union recall = " << maxU << endl;
    if(minrisk) {
      const double mbrH = 100.0 - (100.0*(double)mbr_hamming_sum/hamming_tot);
      cerr << myname << " : min-risk oracle hamming score = "
	   << fixed << setprecision(2) << mbrH << endl;
      assert(mbr_predicted_parts > 0);
      const double mbrP =
	100.0*(double)mbr_matching_parts/(double)mbr_predicted_parts;
      const double mbrR =
	100.0*(double)mbr_matching_parts/(double)correct_parts;
      const double mbrU =
	100.0*(double)mbr_union_recall/(double)correct_parts;
      cerr << myname << " : min-risk oracle prec/rec = "
	   << fixed << setprecision(2)
	   << mbrP << " " << mbrR << endl;
      cerr << myname << " : min-risk union recall = " << mbrU << endl;
    }
  }


  void optutils::load_datasets(structured_model& M,
			       const char* const trfile,
			       dataset& trdata,
			       const char* const valfile,
			       dataset& valdata) {
    int id = 0;
    pattern* x = NULL;
    label* y = NULL;

    assert(trdata.size() == 0);
    assert(valdata.size() == 0);

    /* load the training set */
    cerr << myname << " : loading training set "
	 << fixed << setprecision(1) << flush;
    FILE* in = fopen(trfile, "r");
    assert(in != NULL);
    while(M.fread(in, x, y)) {
      x->set_id(id);
      trdata.push_back(new example(x, y));
      ++id;
      x = NULL;
      y = NULL;
      if((id & 0x7ff) == 0) {
	cerr << "(" << (double)id/1000.0 << "k)" << flush;
      } else if((id & 0xff) == 0) {
	cerr << "." << flush;
      }
    }
    fclose(in);
    cerr << " " << trdata.size() << " examples" << endl;

    /* load validation set if provided */
    if(valfile != NULL) {
      cerr << myname << " : loading validation set "
	   << fixed << setprecision(1) << flush;
      /* NB: we do not reset the id counter; thus all patterns have
	 unique ids. */
      in = fopen(valfile, "r");
      assert(in != NULL);
      int cnt = 0;
      while(M.fread(in, x, y)) {
	x->set_id(id);
	valdata.push_back(new example(x, y));
	++id;
	x = NULL;
	y = NULL;
	++cnt;
	if((cnt & 0x7ff) == 0) {
	  cerr << "(" << cnt/1000 << "k)" << flush;
	} else if((cnt & 0xff) == 0) {
	  cerr << "." << flush;
	}
      }
      fclose(in);
      cerr << " " << valdata.size() << " examples" << endl;
    }
  }

  void optutils::clear_dataset(dataset& data) {
    const int N = data.size();
    for(int i = 0; i < N; ++i) { delete data[i]; }
    data.clear();
  }


  const example* optutils::randselect(const int rand_type,
				      const int i,
				      dataset& data) {
    const int N = data.size();
    /* produce a random value in [0, 1) */
    const double rand_01 = ((double)random())/(1.0 + (double)RAND_MAX);
    if(rand_type == EGSTRA_RAND_UNIFORM) {
      /* in this case, just pick an example at random */
      const double rand_0n = N*rand_01;
      const int rand_idx = (int)rand_0n;
      return data[rand_idx];
    } else if(rand_type == EGSTRA_RAND_PERMUTATION) {
      /* in this case, compute a random permutation on-the-fly by
	 picking a random example out of [i, n) and swapping it
	 with the current example (i.e., the Knuth shuffle). */
      const double rand_in = i + (N - i)*rand_01;
      const int rand_idx = (int)rand_in;
      /* swap and store the random selection in x */
      const example* ex = data[rand_idx];
      data[rand_idx] = data[i];
      data[i] = ex;
      return ex;
    } else if(rand_type == EGSTRA_DETERMINISTIC) {
      /* deterministic ascending order */
      return data[i];
    }

    cerr << "unrecognized randomization type "
	 << rand_type << endl;
    assert(0);

    return NULL;
  }
}

#undef myname
