
/************************************************************
 *
 *   Author : Terry Koo
 *            maestro@csail.mit.edu
 *
 *   Parameter vector class
 *
 ************************************************************/
#ifndef EGSTRA_PARAMETERS_H
#define EGSTRA_PARAMETERS_H

#include <ext/hash_map>  /* for sparse vectors */
#include <string>


#include <assert.h>
#include <math.h>

#include "FVec.h"


  /* cutoff point at which the parameters should be recentered */
#define EGSTRA_PARAMETERS_SCALE_CUTOFF   0.1

namespace egstra {

  /* sparse feature/parameter vector */
  typedef __gnu_cxx::hash_map<int,double> sparsevec;

  /* the parameters are stored internally as c-style array.  the
     _alpha and _Wsq values allow a fast implementation of certain
     operations. */
  class parameters {
  private:
    int _dim;         /* dimensionality */
    double* _W;       /* parameter vector */

    /* fast scaling and addition are implemented by having a separate
       scaling factor _alpha.  specifically, this object represents
       the vector V = _alpha*_W */
    double _alpha;    /* scalar multiplier of _W */
    double _Wsq;      /* squared-L2 of _W */

    /* whether this object owns the memory in _W */
    const bool _self_alloc;

    /* rescale _W so that _alpha = 1.0 and recompute the norm _Wsq */
    void recenter() {
      if(_alpha != 1.0) {
	_Wsq = 0;
	/* rescale and accumulate W^2 */
	for(int i = 0; i < _dim; ++i) {
	  const double d = _alpha*_W[i];
	  _Wsq += d*d;
	  _W[i] = d;
	}
	_alpha = 1.0;
      }
    }
    /* release the arrays */
    void dealloc();

  public:

    /**** lifecycle **********************************************/
    /* create zeroed parameters */
    parameters(const int dim = 0);
    /* create a wrapper around the given vector */
    parameters(double* const V, const int dim);
    /* destroy and delete memory (if _self_alloc) */
    ~parameters();


    /* store/read to file */
    void save(const std::string& fname, const int t,
	      const std::string& stem = "parameters") const;
    void load(const std::string& fname, const int t,
	      const std::string& stem = "parameters");

    /* create a copy of this object */
    parameters* copy() const;

    /* resize the arrays */
    void realloc(const int dim);

    /* accessors */
    int dim() const { return _dim; }

    /* set a batch of values via a sparse vector */
    void set_values(const sparsevec& sV) {
      const double scale = 1.0/_alpha;
      double cur_sqL2 = 0;
      double new_sqL2 = 0;
      sparsevec::const_iterator sV_i = sV.begin();
      const sparsevec::const_iterator sV_end = sV.end();
      for(; sV_i != sV_end; ++sV_i) {
	const int idx = sV_i->first;
	assert(idx >= 0);
	assert(idx < _dim);
	const double cur_val = _W[idx];
	const double new_val = scale*(sV_i->second);
	_W[idx] = new_val;
	/* keep track of the change to w^2 */
	cur_sqL2 += cur_val*cur_val;
	new_sqL2 += new_val*new_val;
      }
      _Wsq += new_sqL2 - cur_sqL2;
    }
    /* copy all values from that to this */
    void set_values(const parameters& that, const double scale = 1.0) {
      _alpha = 1.0;
      _Wsq = 0;
      assert(that._dim == _dim);
      const double* const that_W = that._W;
      const double mul = that._alpha*scale;
      for(int i = 0; i < _dim; ++i) {
	const double val = mul*that_W[i];
	_Wsq += val*val;
	_W[i] = val;
      }
    }


    /**** Read-only computations **********************************/
    /* return the inner product W . F */
    double dot(const struct fvec* const F) const {
      const int n = F->n;
      const int o = F->offset;
      const int* const fidx = F->idx;
      const double* const fval = F->val;
      double iprod = 0;
      if(fval == NULL) { /* indicators */
	for(int i = 0; i < n; ++i) {
	  const int idx = o + fidx[i];
	  assert(idx >= 0);
	  assert(idx < _dim);
	  iprod += _W[idx];
	}
      } else {
	for(int i = 0; i < n; ++i) {
	  const int idx = o + fidx[i];
	  assert(idx >= 0);
	  assert(idx < _dim);
	  iprod += _W[idx]*fval[i];
	}
      }
      return _alpha*iprod;
    }
    /* return the inner product of W with a batch of feature vectors */
    void dot(double* const S, const struct fvec* const F, const int R) const {
      for(int r = 0; r < R; ++r) { S[r] = dot(F + r); }
    }
    /* return the inner product of W with a batch of feature vectors */
    void dot(double* const S, const struct fvec* const F, const int R,
	     const double scale) const {
      for(int r = 0; r < R; ++r) { S[r] = scale*dot(F + r); }
    }
    /* return the inner product W . U */
    double dot(const sparsevec& U) const {
      double iprod = 0;
      sparsevec::const_iterator ui = U.begin();
      const sparsevec::const_iterator uend = U.end();
      for(; ui != uend; ++ui) {
	const int idx = ui->first;
	assert(idx >= 0);
	assert(idx < _dim);
	const double val = ui->second;
	iprod += val*_W[idx];
      }
      return _alpha*iprod;
    }
    /* return W . W */
    double sqL2(const bool force_recenter = false) {
      if(force_recenter) { recenter(); }
      return _alpha*_alpha*_Wsq;
    }

    /**** Modifying computations **********************************/
    /* perform W = 0 */
    void zero() {
      _alpha = 1.0; _Wsq = 0;
      for(int i = 0; i < _dim; ++i) { _W[i] = 0.0; }
    }
    /* perform W *= s using the _alpha variable.  recenter if
       necessary to prevent excessive error */
    void scale(const double s) {
      _alpha *= s;
      if(_alpha < EGSTRA_PARAMETERS_SCALE_CUTOFF ||
	 _alpha > 1.0/EGSTRA_PARAMETERS_SCALE_CUTOFF) {
	recenter();
      }
    }
    /* perform W[i] += d */
    void add(const int i, const double d) {
      assert(i >= 0);
      assert(i < _dim);
      _W[i] += d/_alpha;
    }
    /* perform W += scale * F */
    void add(const struct fvec* const F, const double scale = 1.0) {
      const double realscale = scale/_alpha;
      const int n = F->n;
      const int o = F->offset;
      const int* const fidx = F->idx;
      const double* const fval = F->val;
      double Wsq_adjust = 0;
      if(fval == NULL) { /* indicators */
	const double updsq = realscale*realscale;
	for(int i = 0; i < n; ++i) {
	  const int idx = o + fidx[i];
	  assert(idx >= 0);
	  assert(idx < _dim);
	  const double cur_val = _W[idx];
	  _W[idx] = cur_val + realscale;
	  /* use the formula: (w + x)^2 = w^2 + x^2 + 2*x*w */
	  Wsq_adjust += updsq + 2*realscale*cur_val;
	}
      } else {
	for(int i = 0; i < n; ++i) {
	  const int idx = o + fidx[i];
	  assert(idx >= 0);
	  assert(idx < _dim);
	  const double upd = realscale * fval[i];
	  const double cur_val = _W[idx];
	  _W[idx] = cur_val + upd;
	  /* use the formula: (w + x)^2 = w^2 + x^2 + 2*x*w */
	  Wsq_adjust += upd*upd + 2*upd*cur_val;
	}
      }
      _Wsq += Wsq_adjust;
    }
    /* perform W[k] += scale * U */
    void add(const sparsevec& U, const double scale = 1.0) {
      const double realscale = scale/_alpha;
      double Wsq_adjust = 0;
      sparsevec::const_iterator ui = U.begin();
      const sparsevec::const_iterator uend = U.end();
      for(; ui != uend; ++ui) {
	const int idx = ui->first;
	assert(idx >= 0);
	assert(idx < _dim);
	const double upd = realscale * ui->second;
	const double cur_val = _W[idx];
	_W[idx] = cur_val + upd;
	/* use the formula: (w + x)^2 = w^2 + x^2 + 2*x*w */
	Wsq_adjust += upd*upd + 2*upd*cur_val;
      }
      _Wsq += Wsq_adjust;
    }
    /* perform W *= \exp{scale * U}, where the multiplication is
       component-wise rather than an inner product */
    void expmul(const sparsevec& U, const double scale = 1.0) {
      double Wsq_old = 0;
      double Wsq_new = 0;
      sparsevec::const_iterator ui = U.begin();
      const sparsevec::const_iterator uend = U.end();
      for(; ui != uend; ++ui) {
	const int idx = ui->first;
	assert(idx >= 0);
	assert(idx < _dim);
	const double mul = exp(scale*ui->second);
	const double cur_val = _W[idx];
	const double new_val = mul*cur_val;
	Wsq_old += cur_val*cur_val;
	Wsq_new += new_val*new_val;
	_W[idx] = new_val;
      }
      _Wsq += (Wsq_new - Wsq_old);
    }

    /**** Static utilities **********************************/
    /* find V^2 for sparse vectors */
    static double sparse_sqL2(const sparsevec& V) {
      double norm = 0;
      sparsevec::const_iterator V_i = V.begin();
      const sparsevec::const_iterator V_end = V.end();
      for(; V_i != V_end; ++V_i) {
	const double val = V_i->second;
	norm += val*val;
      }
      return norm;
    }
    /* set V += scale * F for sparse vectors */
    static void sparse_add(sparsevec& V,
			   const struct fvec* const F,
			   const double scale = 1.0) {
      const int n = F->n;
      const int o = F->offset;
      const int* const fidx = F->idx;
      const double* const fval = F->val;
      if(fval == NULL) { /* indicators */
	for(int i = 0; i < n; ++i) {
	  const int idx = o + fidx[i];
	  V[idx] += scale;
	}
      } else {
	for(int i = 0; i < n; ++i) {
	  const int idx = o + fidx[i];
	  const double update = scale*(fval[i]);
	  V[idx] += update;
	}
      }
    }

    /* array access: these involve multiplication and division of
       doubles, so don't use them often */
    double operator[](const int i) const {
      assert(i >= 0);
      assert(i < _dim);
      return _alpha*_W[i];
    }

    /* direct access to the parameter vector.  use only if you know
       what you're doing. */
    double* c_buf() { assert(_W != NULL); return _W; }

  };
}

#endif /* EGSTRA_PARAMETERS_H */
