/* StructuredModel : abstract class representing a part-factorized
   structured model

   Create a subclass of this class and define whichever of the
   inference routines you need.  To classify data, you will need to
   define the argmax() routine.  To use EG, you will need to define
   the marginals() routine.  To train log-linear models in the primal,
   you will need to define the marginals() and partition() routines.

   The lossvec() should return a vector of indicators that indicates
   which parts are "incorrect" according to a Hamming loss criterion.
   A default implementation is provided which simply uses the label to
   define the Hamming loss directly; however, you may wish to redefine
   it depending on the nature of your model.  Many structured models
   are defined at a "higher" order than the level at which they are
   scored; for example, a trigram part-of-speech tagger defines parts
   made up of trigrams of tags, but the tagger's output will be scored
   at the level of tag unigrams, not tag trigrams.  If you are
   training a max-margin Markov network, it would be sensible to
   define the Hamming loss only in terms of the tag unigrams instead
   of the tag trigrams, so that the model is optimized for the scoring
   metric.

   It is important to have a consistent method for numbering the parts
   in a given pattern.

   You should redefine the _name to be descriptive of the subclass,
   for debug-output purposes.

   The usage() and process_options() functions allow the model to
   handle command-line options (see Options.h/cc).

   Author : Terry Koo
            maestro@csail.mit.edu */

#ifndef EGSTRA_STRUCTURED_MODEL_H
#define EGSTRA_STRUCTURED_MODEL_H

#include "Example.h"
#include "Pattern.h"
#include "Label.h"

#include <stdio.h>

namespace egstra {
  class structured_model {
  protected:
    const char* _name;

  public:
    structured_model(const char* name) : _name(name) {}
    virtual ~structured_model() {}

    /* handle the command line */
    virtual void usage(const char* const mesg = "") = 0;
    virtual void process_options() = 0;

    /* accessors */
    const char* name() { return _name; }

    /* read the next (x_i,y_i) pair out of the provided data file.  if
       the are no more pairs, then return false, otherwise set the
       values of x and y and return true. */
    virtual bool fread(FILE* const in, pattern*& x, label*& y) = 0;
    /* write the (x,y) pair to the provided file.  this routine will
       be used to save decoded labels for held-out validation data. */
    virtual void fwrite(FILE* const out, const pattern* const x,
			const label& y) = 0;


    /* The indicator vector should be returned as a char* array that
       is allocated by malloc(). */
    virtual char* lossvec(const example* const ex);

    virtual double argmax(const pattern* const x, const double* const S,
			  label& y);
    virtual void kbest_init(const pattern* const x, const double* const S);
    virtual void kbest_clear();
    virtual bool kbest(const int k, const pattern* const x,
		       const double* const S, label& y, double& score);
    virtual double partition(const pattern* const x, const double* const S);
    virtual double marginals(const pattern* const x, const double* const S,
			     double* const M);
  };
}

#endif /* EGSTRA_STRUCTURED_MODEL_H */
