/************************************************************
 *
 *   Author : Xavier Carreras and Terry Koo
 *            {carreras,maestro}@csail.mit.edu
 *
 *   Randomized Online(-like) Dual Exponentiated Gradient Training
 *
 ************************************************************/
#ifndef EGSTRA_DEGTRAINER_H
#define EGSTRA_DEGTRAINER_H

#include "Trainer.h"
#include "Parameters.h"
#include "DualVector.h"

#include <string>

namespace egstra {
  /**
   * DEGTrainer : a templatized class dependent on an objective
   * function Obj
   *
   * Obj is expected to provide the following meathods
   *
   *   Obj::exprimal(example* ex, double* S, structured_model M)
   *   Obj::exgrad(parameters g, double scale, example* ex,
   *               double* S, fvec* F, structured_model M)
   *     These return the example-wise contribution to the primal
   *     objective and the primal gradient
   *
   *   Obj::update_dual_variable(theta, iprod, loss, scale)
   *     Take the current dual variable and other information and
   *     return an updated dual variable.
   *
   *   Obj::exdual_init(logZ)
   *   Obj::exdual_contrib(mu, theta, loss)
   *     These allow computation of the dual objective (decomposed
   *     into two functions)
   *
   *   Obj::lrate_search_init()
   *     Return a starting point for the estimated learning-rate
   *     search.
   **/
  template<class Obj>
  class deg_trainer : public trainer {
  private:
    /**** global algorithm parameters *****/
    /* regularization constant */
    double _C;
    /* global initial learning rate (eta0) */
    double _lrate;
    /* initial bias on the true edges */
    double _beta;
    /* primal parameters */
    parameters _W;
    /* structured dual variables (log-scores and marginals) */
    dual_vector* _theta;
    dual_vector* _mu;
    /* current log-partition for all examples */
    double* _logZ;

    /***** global settings *****/
    /* type of randomization (see above) */
    int _rand_type;
    /* seed for srandom() for randomized online training */
    int _seed;
    /* whether the learning rate is per-example adaptive */
    double _adapt;
    /* disk cache for storing dual variables (memory cache if empty) */
    std::string _dualcache;
    /* whether the dual cache is gzipped */
    bool _dualgzipped;
    /* whether to write parameter vectors to disk */
    bool _writeparams;
    /* whether to compute the primal objective on each iteration */
    bool _doprimal;
    /* whether to compute a regularization path */
    bool _regpath;
    /* gap tolerance for regularization path */
    double _regpath_tol;
    /* step size for regularization path */
    double _regpath_step;

    /***** training invocation variables *****/
    /* total number of training examples visited */
    int _num_processed;
    /* total amount of time spent optimizing */
    double _cpu_time;
    /* this is equal to 1/(N*C), where N is the size of the training
       set.  this value has significance in that the primal-dual
       correspondence is defined as

       W(A) = 1/(N*C) \sum_iy a_iy psi_iy

       (note that this differs from the paper, due to a factor of 1/N
       in the primal objective)*/
    double _invNC;
    /* number of training examples */
    int _N;

    /***** per-example quantities *****/
    /* per-example duals (e.g., negative entropy, expected loss) */
    double* _exduals;
    /* separate learning rate for every example */
    double* _lrates;


    /* given a sentence, attempt to update its duals, halving the
       learning rate if necessary, and updating the primals and dual
       objective value.  t is the iteration-number. */
    void process_example(const example* const ex,
			 const struct fvec* const F,
			 const int t);

    /* initialize the duals and update the primals */
    void initialize_example(const example* const ex,
			    const struct fvec* const F);

    /* given the training set, pick a small sample of sentences and
       attempt to find a good learning rate */
    double estimate_lrate(const dataset& data);
    /* helper function for the above */
    int count_cuts(const dataset& sample, const double lrate);

    /* reconstruct the primal vector */
    void reconstruct_primals(const dataset& data);

    /* helper that (re-)initializes the dual-vectors */
    void initialize_duals();

  public:

    deg_trainer()
      : trainer(), _theta(NULL), _mu(NULL), _logZ(NULL),
        _exduals(NULL), _lrates(NULL) {}
    ~deg_trainer();

    void usage(const char* const mesg = "");
    void process_options();

    void train(dataset& trdata, dataset& valdata, const int T);

  };

}

#include "DEGTrainer.tcc"


#endif /* EGSTRA_DEGTRAINER_H */
