#include "Decoder.h"

#include "Options.h"
#include "OptUtils.h"
#include "Parameters.h"

#include <stdlib.h>

#include <iostream>
#include <string>

using namespace std;

namespace egstra {
  void decoder::usage(const char* const mesg) {
    cerr << "Decoder options:" << endl;
    cerr << " --minrisk        : use min-risk decoding (max-marginals)" << endl;
    cerr << " --kbest=<int>    : use k-best decoding with the given k" << endl;
    cerr << " --range=<int>... : list of iteration numbers indicating which" << endl;
    cerr << "                    parameters to use when decoding; e.g.," << endl;
    cerr << "                    --range=2,3,5,7" << endl;
    cerr << "                    --range=1-100" << endl;
    cerr << "                    --range=20-40,1,2,99-200" << endl;
    cerr << endl;
    cerr << mesg << endl;
  }

  void decoder::process_options() {
    int tmp;
    _minrisk = false;
    if(options::get("minrisk", tmp)) {
      _minrisk = true;
    }

    _kbest = 0;
    if(options::get("kbest", tmp)) {
      _kbest = tmp;
      if(_kbest <= 0) {
	usage("please provide a k > 0"); exit(1);
      }
    }

    string s;
    if(!options::get("range", s)) {
      usage("please provide a range of iterations"); exit(1);
    }

    const char* cs = s.c_str();
    char* str = new char[s.size() + 1];
    for(int i = 0; i < (int)s.size(); ++i) {
      str[i] = cs[i];
    }
    str[s.size()] = '\0';
    char* p = str;
    while(1) {
      const int next = strtol(p, &p, 10);
      if(*p == ',' || *p == '\0') {
	_range.push_back(next);
      } else if(*p == '-') {
	++p;
	const int last = strtol(p, &p, 10);
	if((*p != ',' && *p != '\0') || last < next) {
	  cerr << "malformed range: \"" << s << "\"" << endl;
	  exit(1);
	}
	for(int i = next; i <= last; ++i) {
	  _range.push_back(i);
	}
      } else {
	cerr << "malformed range: \"" << s << "\"" << endl;
	exit(1);
      }
      if(*p == '\0') { break; }
      ++p;
    }
    delete [] str;
  }

  void decoder::train(dataset& trdata, dataset& valdata,
		      const int T) {
    cerr << "Decoder : using parameters";
    for(int i = 0; i < (int)_range.size(); ++i) {
      cerr << " " << _range[i];
    }
    cerr << endl;
    parameters w;
    for(int i = 0; i < (int)_range.size(); ++i) {
      const int t = _range[i];
      w.load(_dir, t);
      assert(w.dim() == _fg->dim());
      if(_kbest == 0) {
	optutils::classify_val_set(w, trdata, *_fg, *_M,
				   _dir, t, _minrisk);
      } else {
	optutils::kclassify_val_set(w, trdata, *_fg, *_M,
				    _dir, t, _kbest, _minrisk);
      }
    }
  }
}
