#include "NullTrainer.h"

#include "Options.h"
#include "Timer.h"

#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <assert.h>
#include <limits.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/wait.h>

#include <iostream>
#include <sstream>

#define MAX_NUM_WORKERS     4
#define MAX_WORKER_RETRIES  5

using namespace std;

namespace egstra {
  static pid_t safe_fork(const char* mesg) {
    const pid_t p = fork();
    if(p == -1) {
      perror(mesg); assert(0);
    } else if(p < 0) {
      cerr << mesg << ": unhandled retval from fork(): " << p << endl;
      assert(0);
    }
    return p;
  }

  /* print something to stderr while buffering it in a string
     stream */
#define bufprint(txt)		      \
  {				      \
    ostringstream sbuf;		      \
    sbuf << txt;		      \
    cerr << sbuf.str() << flush;      \
  }
#define bufprintln(txt)		      \
  {				      \
    ostringstream sbuf;		      \
    sbuf << txt;		      \
    cerr << sbuf.str() << endl;	      \
  }


  static pid_t worker(const int read_fd, const int write_fd,
		      const int wid, fgen* const fg,
		      const dataset& trdata, const dataset& valdata) {
    ostringstream os;
    os << "worker " << wid;
    const string name(os.str());
    const pid_t p = safe_fork(name.c_str());
    if(p > 0) { return p; }
    assert(close(write_fd) == 0);
    bufprintln("worker " << wid << ": initialized");
    int t;
    int retries = 0;
    while(1) {
      const ssize_t size = read(read_fd, &t, sizeof(int));
      if(size == sizeof(int)) {
	retries = 0;
      } else if(size == 0) {
	bufprintln(name << ": job queue finished");
	exit(0);
      } else if(size == -1) {
	if(errno == EAGAIN) {
	  ++retries;
	  bufprintln(name << ": no jobs (attempt " << retries << ")");
	  if(retries >= MAX_WORKER_RETRIES) {
	    bufprintln(name << ": warning, no jobs after "
		       << retries << " attempts, exiting");
	    exit(0);
	  }
	  sleep(1 + (rand() % 3));
	  continue;
	}
	bufprintln(name << ": *** error reading from pipe");
	perror(name.c_str()); assert(0);
      } else {
	bufprintln(name << ": unexpected retval from read(): " << size);
	assert(0);
      }

      /* handle the provided example */
      const int ntr = trdata.size();
      const example* const ex = (t < ntr ? trdata[t] : valdata[t - ntr]);
      const pattern* const x = ex->x();
      if(t != x->id()) {
	bufprintln(name << ": index mismatch: t=" << t
		   << " xid=" << x->id());
	assert(0);
      }
      fg->discard(fg->phi(x), x);
      if(((t + 1) & 0xff) == 0) {
	bufprint("(" << fixed << setprecision(1)
		 << (double)t/1000.0 << "k)");
      } else if(((t + 1) & 0x1f) == 0) {
	cerr << "." << flush;
      }
    }
    cerr << "worker " << wid << ": *** reached end of function" << endl;
    assert(0);
    return p;
  }

  static pid_t manager(const int read_fd, const int write_fd,
		       const int beg, const int end,
		       const dataset& trdata, const dataset& valdata) {
    const pid_t p = safe_fork("manager");
    if(p > 0) { return p; }
    assert(close(read_fd) == 0);
    bufprintln("manager: initialized");
    /* unload training data indices */
    const int ntr = trdata.size();
    int i;
    for(i = beg; i < end && i < ntr; ++i) {
      const int t = trdata[i]->x()->id();
      assert(i == t);
      const ssize_t size = write(write_fd, &t, sizeof(int));
      if(size == -1) {
	perror("manager"); assert(0);
      } else if(size != sizeof(int)) {
	bufprintln("manager: unexpected retval " << size);
	assert(0);
      }
    }
    /* unload validation data indices */
    const int val_end = ntr + valdata.size();
    for(; i < end && i < val_end; ++i) {
      const int t = valdata[i - ntr]->x()->id();
      assert(i == t);
      const ssize_t size = write(write_fd, &t, sizeof(int));
      if(size == -1) {
	perror("manager"); assert(0);
      } else if(size != sizeof(int)) {
	bufprintln("manager: unexpected retval " << size);
	assert(0);
      }
    }
    bufprintln("manager: finished");
    assert(close(write_fd) == 0);
    exit(0);
    return p;
  }



  void null_trainer::usage(const char* const mesg) {
    cerr << "NullTrainer options:" << endl;
    cerr << " --workers=<int>     : number of processes to use to enumerate features" << endl;
    cerr << " --range=<beg>-<end> : range of examples to process" << endl;
    cerr << endl;
    cerr << mesg << endl;
  }

  void null_trainer::process_options() {
    int tmp;
    _njobs = -1;
    if(options::get("workers", tmp)) {
      if(tmp > MAX_NUM_WORKERS) {
	usage("can't use that many workers"); exit(1);
      } else if(tmp <= 0) {
	usage("number of workers must be > 0"); exit(1);
      }
      _njobs = tmp;
    }

    string s;
    _range_beg = 0;
    _range_end = INT_MAX;
    if(options::get("range", s)) {
      const char* const str = s.c_str();
      char* const arr = new char[s.size() + 1];
      for(int i = 0; i <= (int)s.size(); ++i) { arr[i] = str[i]; }
      char* p = arr;
      const int beg = strtol(p, &p, 10);
      if(*p != '-')  { usage("malformed range"); exit(1); }
      ++p;
      const int end = strtol(p, &p, 10);
      if(*p != '\0') { usage("malformed range"); exit(1); }
      if(beg < 0)    { usage("malformed range"); exit(1); }
      if(end < beg)  { usage("malformed range"); exit(1); }
      _range_beg = beg;
      _range_end = end + 1;
      delete [] arr;
    }
  }

  void null_trainer::train(dataset& trdata, dataset& valdata,
			   const int T) {
    if(_njobs == -1) { /* normal single-threaded version */
      cerr << "NullTrainer : generating training features " << flush;
      timer_start();
      const int Ntr = (_range_end < (int)trdata.size()
		       ? _range_end : (int)trdata.size()) - _range_beg;
      for(int i = 0; i < Ntr; ++i) {
	const example* const ex = trdata[i + _range_beg];
	_fg->discard(_fg->phi(ex->x()), ex->x());
	ticker(i + 1, Ntr);
      }
      ticker(Ntr, Ntr);
      timer_stop();
      cerr << " [" << timestr(timediff) << "]" << endl;

      if(valdata.size() > 0) {
	cerr << "NullTrainer : generating test features (make sure you don't do this when generating a feature dictionary) " << flush;
	timer_start();
	const int Nval = (_range_end < (int)valdata.size()
			  ? _range_end : (int)valdata.size()) - _range_beg;
	for(int i = 0; i < Nval; ++i) {
	  const example* const ex = valdata[i + _range_beg];
	  _fg->discard(_fg->phi(ex->x()), ex->x());
	  ticker(i + 1, Nval);
	}
	ticker(Nval, Nval);
	timer_stop();
	cerr << " [" << timestr(timediff) << "]" << endl;
      }
    } else { /* multi-process version */

      /* the training is broken into a manager child and worker
	 children, which all communicate using a pipe */
      int fds[2];
      const int ret = pipe(fds);
      if(ret != 0) {
	cerr << "NullTrainer : pipe() failed" << endl;
	perror("pipe"); assert(0);
      }
      const int pipe_rd = fds[0];
      const int pipe_wr = fds[1];
      pid_t* const child_pids = new pid_t[_njobs + 1];
      /* set non-blocking I/O for the pipe */
      assert(fcntl(pipe_rd, F_SETFL, O_NONBLOCK) != -1);

      /* start up the manager and workers */
      timer_start();
      cerr << "NullTrainer : generating features with "
	   << _njobs << " workers" << endl;
      cerr << fixed << setprecision(1);
      child_pids[0] = manager(pipe_rd, pipe_wr, _range_beg, _range_end,
			      trdata, valdata);
      for(int i = 1; i <= _njobs; ++i) {
	child_pids[i] = worker(pipe_rd, pipe_wr, i, _fg, trdata, valdata);
      }

      /* the pipe is unused by this process */
      assert(close(pipe_wr) == 0);
      assert(close(pipe_rd) == 0);

      /* note each child as it terminates, and set failure if it
	 exited abnormally or was killed */
      int failure = 0;
      int nchildren;
      for(nchildren = _njobs + 1; nchildren > 0; --nchildren) {
	int status;
	const pid_t child = wait(&status);
	if(child == -1) {
	  cerr << "NullTrainer : wait() failed" << endl;
	  perror("wait"); assert(0);
	} else {
	  /* attempt to find the child matching the PID */
	  int idx = -1;
	  for(int i = 0; i <= _njobs; ++i) {
	    if(child == child_pids[i]) { idx = i; break; }
	  }
	  if(idx ==-1) {
	    cerr << "NullTrainer : unknown child pid " << child << endl;
	    assert(0);
	  }
	  ostringstream os;
	  if(idx == 0) { os << "NullTrainer : manager "; }
	  else         { os << "NullTrainer : worker " << idx << " "; }
	  if(WIFEXITED(status)) {
	    const int estat = WEXITSTATUS(status);
	    if(estat != 0) {
	      failure = 1;
	      cerr << os.str() << "exited with status " << estat << endl;
	    } else {
	      cerr << os.str() << "exited normally" << endl;
	    }
	  } else if(WIFSIGNALED(status)) {
	    const int sig = WTERMSIG(status);
	    failure = 1; /* death by signal is a failure */
	    cerr << os.str() << "killed by signal " << sig << endl;
	  } else {
	    failure = 1; /* consider this a failure */
	    cerr << os.str() << "unknown status " << status << endl;
	  }
	}
      }
      if(failure) {
	cerr << "NullTrainer : detected failure in a child" << endl;
	assert(0);
      }

      delete [] child_pids;

      timer_stop();
      cerr << "NullTrainer : elapsed time ["
	   << timestr(timediff) << "]" << endl;
    }
  }
}
