#include "PercTrainer.h"

/**
 * Implements a primal perceptron trainer
 * @file PercTrainer.cc
 * @author Terry Koo, Xavier Carreras
 **/

#include "Timer.h"
#include "Options.h"
#include "Constants.h"
#include "OptUtils.h"

#include <iostream>
#include <fstream>
#include <iomanip>

#include <assert.h>
#include <stdlib.h>
#include <string.h>

/* detailed name of the trainer */
#define myname							\
  (_averaged ? "A" : "")					\
  << RAND_SHORT_GLOSS(_rand_type)				\
  << "PT[" << _M->name() << "]"

using namespace std;

namespace egstra {
  void perc_trainer::usage(const char* const mesg) {
    cerr << "PercTrainer options:" << endl;
    cerr << " --rand=<string> : randomization type (det,unif,perm)" << endl;
    cerr << " --seed=<int>    : random seed" << endl;
    cerr << " --averaged      : average the weight vectors" << endl;
    cerr << " --margin=<real> : margin-sensitivity (default 0.0)" << endl;
    cerr << " --noparams      : do not write weight vectors to disk" << endl;
    cerr << " --partwise      : enumerate features part-by-part instead of" << endl;
    cerr << "                   grabbing all parts at once" << endl;
    cerr << endl;
    cerr << mesg << endl;
  }

  void perc_trainer::process_options() {
    _rand_type = EGSTRA_RAND_UNIFORM;
    string rtype;
    if(options::get("rand", rtype)) {
      if(strcmp(rtype.c_str(), "det") == 0) {
	_rand_type = EGSTRA_DETERMINISTIC;
      } else if(strcmp(rtype.c_str(), "unif") == 0) {
	_rand_type = EGSTRA_RAND_UNIFORM;
      } else if(strcmp(rtype.c_str(), "perm") == 0) {
	_rand_type = EGSTRA_RAND_PERMUTATION;
      } else {
	usage("unrecognized randomization type"); exit(1);
      }
    }

    int tmp;
    _seed = 1;
    if(options::get("seed", tmp)) {
      _seed = tmp;
    }

    _averaged = false;
    if(options::get("averaged", tmp)) {
      _averaged = true;
    }

    _margin = 0;
    double margin;
    if(options::get("margin", margin)) {
      usage("sorry, the margin perceptron is not implemented"); exit(1);
      _margin = margin;
    }

    _writeparams = true;
    if(options::get("noparams", tmp)) {
      _writeparams = false;
    }

    _partwise = false;
    if(options::get("partwise", tmp)) {
      _partwise = true;
    }
  }


  /* adjust the primals with the gradient on the given example */
  void perc_trainer::process_example(const example* const ex) {
    const pattern* x = ex->x();
    const label& y = ex->y();
    const int R = x->nparts();

    /* get the inner products */
    double* const S = (double*)malloc(R*sizeof(double));
    const struct fvec* F = NULL;
    if(_partwise) {
      for(int r = 0; r < R; ++r) {
	const struct fvec* const f_r = _fg->phi(x, r);
	S[r] = _W.dot(f_r);
	_fg->discard(f_r, x, r);
      }
    } else {
      F = _fg->phi(x);
      _W.dot(S, F, R);
    }

    /* find the argmax */
    label yhat;
    _M->argmax(x, S, yhat);
    free(S);

    /* collect the part-wise contributions (using char is okay since
       the contrib will only ever be -1, 0, +1) */
    char* const contrib = (char*)calloc(R, sizeof(char));
    { /* true parts receive +1 weight */
      label::const_iterator it = y.begin();
      const label::const_iterator it_end = y.end();
      for(; it != it_end; ++it) {
	const int part = *it;
	assert(part >= 0);
	assert(part < R);
	++(contrib[part]);
      }
    }
    { /* predicted parts receive -1 weight */
      label::const_iterator it = yhat.begin();
      const label::const_iterator it_end = yhat.end();
      for(; it != it_end; ++it) {
	const int part = *it;
	assert(part >= 0);
	assert(part < R);
	--(contrib[part]);
      }
    }

    /* write the gradient update into a sparse vector */
    bool mistake = false;
    sparsevec update;
    if(_partwise) {
      for(int r = 0; r < R; ++r) {
	if(contrib[r] != 0) {
	  const struct fvec* const f_r = _fg->phi(x, r);
	  parameters::sparse_add(update, f_r, contrib[r]);
	  _fg->discard(f_r, x, r);
	  mistake = true;
	}
      }
    } else {
      for(int r = 0; r < R; ++r) {
	if(contrib[r] != 0) {
	  parameters::sparse_add(update, F + r, contrib[r]);
	  mistake = true;
	}
      }
    }
    if(mistake) { ++_mistakes; }

    if(!_partwise) { _fg->discard(F, x); }
    free(contrib);

    /* update the sum vector if necessary */
    if(_averaged) {
      /* bring the _Wsum vector up to date with the state just before
	 this update */
      const int now = _num_processed;
      const double* const wbuf = _W.c_buf();
      sparsevec::const_iterator ui = update.begin();
      const sparsevec::const_iterator uend = update.end();
      for(; ui != uend; ++ui) {
	const int idx = ui->first;
	const double cur_val = wbuf[idx];
	const int time = _Wtime[idx];
	const int elapsed = now - time;
	_Wsum[idx] += elapsed*cur_val;
	_Wtime[idx] = now;
      }
    }

    /* update the primals */
    _W.add(update);
  }

  /* perform all lazy updates on the sum-vector */
  void perc_trainer::flush_avg() {
    const int now = _num_processed;
    const int d = _W.dim();
    const double* const wbuf = _W.c_buf();
    for(int i = 0; i < d; ++i) {
      _Wsum[i] += (now - _Wtime[i])*wbuf[i];
      _Wtime[i] = now;
    }
  }

  void perc_trainer::train(dataset& trdata, dataset& valdata,
			   const int T) {
    _W.realloc(_fg->dim());

    cerr << myname << " : training started for max " << T
	 << " rounds ... " << endl;
    cerr << myname << " : data order is "
	 << RAND_LONG_GLOSS(_rand_type) << endl;

    if(_rand_type != EGSTRA_DETERMINISTIC) {
      /* only report this for randomized EG */
      cerr << myname << " : seeding the RNG with value " << _seed << endl;
      srandom(_seed);
    }

    const int Ntr = trdata.size();

    /* log file for primal and #pass information */
    string objfilename = _dir + "/obj.out";
    ofstream obj_out(objfilename.c_str(), ios::out | ios::trunc);
    assert(obj_out.good());

    _W.zero();
    if(_averaged) {
      /* allocate vectors for avg-params and timestamps */
      cerr << myname << " : allocating averaged vectors...";
      const int d = _W.dim();
      _Wsum = new double[d];
      _Wtime = new int[d];
      for(int i = 0; i < d; ++i) { _Wsum[i] = 0.0; _Wtime[i] = 0; }
      cerr << " done" << endl;
    }

    _num_processed = 0;
    _mistakes = 0;
    double cpu_time = 0;
    for(int t = 1; t <= T; ++t) {
      cerr << myname << " : iter " << t << " " << flush;
      timer_start();
      const clock_t t0 = clock();
      for(int i = 0; i < Ntr; ++i) {
	/* first, pick an example according to the desired scheme.
	   NB: the first round of rand-uniform will be a permutation,
	   to ensure that all examples are visited. */
	const int rt = (t == 1 && _rand_type == EGSTRA_RAND_UNIFORM
			? EGSTRA_RAND_PERMUTATION : _rand_type);
	const example* const ex = optutils::randselect(rt, i, trdata);

	/* perform a perceptron update */
	process_example(ex);
	++_num_processed;

	ticker(i + 1, Ntr);
      }
      ticker(Ntr, Ntr);
      timer_stop();
      cpu_time += ((double)(clock() - t0))/(double)CLOCKS_PER_SEC;
      cerr << " [" << timestr(timediff) << "]" << endl;

      /* save the primals in the model directory */
      const parameters* _Wavg = NULL;
      if(_averaged) {
	flush_avg();
	_Wavg = new parameters(_Wsum, _W.dim());
	if(_writeparams) { _Wavg->save(_dir, t); }
      } else {
	if(_writeparams) { _W.save(_dir, t); }
      }

      /* #passes is an abstract measure of computation */
      const double num_passes = ((double)_num_processed)/((double)Ntr);
      /* update the objective log with the current iteration */
      obj_out << scientific << setprecision(16)
	      << t << " " << num_passes << " "
	      << cpu_time << " "
	      << _W.sqL2() << " "
	      << _mistakes << endl;
      /* print a summary */
      cerr << myname << " : finished it = " << t
	   << scientific << setprecision(5)
	   << " ; ops = " << num_passes
	   << " ; cpu = " << cpu_time
	   << " ; W^2 = " << _W.sqL2()
	   << " ; mistakes = " << _mistakes <<  endl;

      /* parse the validation data if provided */
      if(valdata.size() > 0) {
	/* NB: the perceptron is not a probabilistic learning
	   approach, so it does not make sense to compute minimum-risk
	   labels */
	if(_averaged) {
	  optutils::classify_val_set(*_Wavg, valdata,
				     *_fg, *_M, _dir, t, false);
	} else {
	  optutils::classify_val_set(_W, valdata,
				     *_fg, *_M, _dir, t, false);
	}
      }

      /* clean up temporary averaged parameters */
      if(_averaged) { delete _Wavg; }
    }


    /* tidy up */
    if(_averaged) {
      delete [] _Wsum; _Wsum = NULL;
      delete [] _Wtime; _Wtime = NULL;
    }
    obj_out.close();
  }
}

#undef myname
