/* Registry : A class that provides facilities for dynamically
   allocating modules like optimizers (Trainer), feature generators
   (FGen), and structured models (StructuredModel).  This is based on
   the notion of auto-registration of objects.  See also Register.h.

   Author : Terry Koo
            maestro@csail.mit.edu */
#ifndef EGSTRA_REGISTRY_H
#define EGSTRA_REGISTRY_H

#include <map>
#include <string>
#include <vector>

namespace egstra {
  /* pointer to a function that creates an instance of a class */
  typedef void* (*constructor)();
  /* map from class-name to constructor */
  typedef std::map<std::string,constructor> consmap;
  /* map from type-name to consmap */
  typedef std::map<std::string,consmap> reg;

  class registry {
  private:
    /* global mapping from class-type to mapping from class-names to
       constructors.  this must be a pointer so that it can be
       properly initialized after the program starts. */
    static reg* _registry;

  public:
    /* instantiating a Registry object registers a constructor.  this
       is convenient for the purposes of auto-registering classes. */
    registry(const char* type, const char* name, constructor f);
    ~registry();

    /* register a constructor for the given type and name.  it is an
       error to register something for a type/name combination that is
       already registered */
    static void add(const char* type, const char* name, constructor f);

    /* create an object of the given type and name and return it */
    static void* construct(const char* type, const char* name);

    /* get a list of registered class-names of the given type */
    static void getnames(const char* type, std::vector<std::string>& names);
  };
}

#endif /* EGSTRA_REGISTRY_H */
